import sublime, sublime_plugin
import subprocess
import re
import os
import os.path as osp
from time import *

SETTINGS_FILE = "AutoIt.sublime-settings"
settings = None

class CompileAu3Command(sublime_plugin.WindowCommand):
    def run(self):
        compilepath = settings.get("autoit_compiler_path", None)
        if not compilepath:
            return
        else:
            compilepath = '"%s"' % compilepath

        view = self.window.active_view()
        curr_full_path = view.file_name()
        curr_folder = osp.dirname(curr_full_path)
        curr_file_ext = osp.basename(curr_full_path)
        curr_file, ext = osp.splitext(curr_file_ext)
        if not re.search("[Aa][Uu]3", ext):
            return
        patt_outfile = "^#AutoIt3Wrapper_OutFile=(.*)"
        patt_icon = "^#AutoIt3Wrapper_Icon=(.*)"
        patt_compression = "^#AutoIt3Wrapper_Compression=(.*)"
        patt_upx = "^#AutoIt3Wrapper_UseUpx=(.*)"
        patt_console = "^#AutoIt3Wrapper_Change2CUI=(.*)"

        infile = ' /in "%s"' % curr_full_path
        outfile = ' /out "%s"' % osp.join(curr_folder, "%s.exe" %  curr_file)
        icon = ""
        compression = " /comp 0"
        upx = " /pack"
        console = " /gui"

        match = view.find(patt_outfile, 0, sublime.IGNORECASE)
        if match:
            inmatch = re.search("(=)(.*)", view.substr(match))
            outfile = ' /out "%s"' % inmatch.group(2)
            path, filename = osp.split(inmatch.group(2))
            if not osp.exists(path):
                os.makedirs(path)

        match = view.find(patt_icon, 0, sublime.IGNORECASE)
        if match:
            inmatch = re.search("(=)(.*)", view.substr(match))
            icon = ' /icon "%s"' % inmatch.group(2)

        match = view.find(patt_compression, 0, sublime.IGNORECASE)
        if match:
            inmatch = re.search("(=)(.*)", view.substr(match))
            if re.search("[0-4]", inmatch.group(2)):
                compression = " /comp %s" % inmatch.group(2)

        match = view.find(patt_upx, 0, sublime.IGNORECASE)
        if match:
            inmatch = re.search("(=)(.*)", view.substr(match))
            if re.search("[nN]", inmatch.group(2)):
                upx = " /nopack"

        match = view.find(patt_console, 0, sublime.IGNORECASE)
        if match:
            inmatch = re.search("(=)(.*)", view.substr(match))
            if re.search("[yY]", inmatch.group(2)):
                console = " /console"

        str_compile = "%s%s%s%s%s%s%s" % (compilepath, infile, outfile, icon,
            compression, upx, console)

        print('\n'*10)
        print("------------ Beginning Compile ------------ %s" % time_now())
        print("cmd: %s" % str_compile)
        self.window.run_command("exec", {"cmd": str_compile})
        print("-------------- Compile Ended -------------- %s" % time_now())


class WrapperDirectives(sublime_plugin.TextCommand):
    window = sublime.active_window()
    directives = "#Region AutoIt3Wrapper directives\n"
    outpath = ""
    icon = ""
    upx = "y"
    compression = "2"
    console = "n"

    def run(self, edit):
        wrapper_defaults = settings.get("wrapper_defaults", None)
        if wrapper_defaults:
            self.outpath = wrapper_defaults.get("out_path", "")
            self.icon = wrapper_defaults.get("icon", "")
            self.upx = wrapper_defaults.get("upx", "y")
            self.compression = wrapper_defaults.get("compression", "2")
            self.console = wrapper_defaults.get("console", "n")
        curr_full_path = self.window.active_view().file_name()
        curr_folder = osp.dirname(curr_full_path)
        curr_file_ext = osp.basename(curr_full_path)
        curr_file, ext = osp.splitext(curr_file_ext)
        if not re.search("[Aa][Uu]3", ext):
            return

        if self.outpath != "":
            curr_folder = self.outpath
        self.get_outfile("%s" % osp.join(curr_folder, "%s.exe" %  curr_file))

    def get_outfile(self, fileout):
        self.window.show_input_panel("#AutoIt3Wrapper_OutFile = ",
            fileout, self.on_outfile, None, None)

    def on_outfile(self, user_input):
        if user_input != '':
            self.directives = ('%s#AutoIt3Wrapper_OutFile=%s\n' % (
                self.directives, user_input))
        self.get_icon()

    def get_icon(self):
        self.window.show_input_panel("#AutoIt3Wrapper_Icon = ", self.icon,
            self.on_icon, None, None)

    def on_icon(self, user_input):
        if user_input != '':
            self.directives = ('%s#AutoIt3Wrapper_Icon=%s\n' % (
                self.directives, user_input))
        self.get_upx()

    def get_upx(self):
        self.window.show_input_panel("#AutoIt3Wrapper_UseUpx = ", self.upx,
            self.on_upx, None, None)

    def on_upx(self, user_input):
        if user_input == 'y' or user_input == 'Y':
            self.directives = ('%s#AutoIt3Wrapper_UseUpx=%s\n' % (
                self.directives, user_input[:1]))
            self.get_compression()
        else:
            self.get_cui()

    def get_compression(self):
        self.window.show_input_panel("#AutoIt3Wrapper_Compression = ",
            self.compression, self.on_compression, None, None)

    def on_compression(self, user_input):
        if user_input != '':
            self.directives = ('%s#AutoIt3Wrapper_Compression=%s\n' % (
                self.directives, user_input[:1]))
        self.get_cui()

    def get_cui(self):
        self.window.show_input_panel("#AutoIt3Wrapper_Change2CUI = ",
            self.console, self.on_cui, None, None)

    def on_cui(self, user_input):
        if user_input == 'y' or user_input == 'Y':
            self.directives = ('%s#AutoIt3Wrapper_Change2CUI=%s\n' % (
                self.directives, user_input[:1]))
        self.ask_insertpos()

    def ask_insertpos(self):
        self.window.show_input_panel("Line to insert Wrapper directives ", '1',
            self.on_insert, None, None)

    def on_insert(self, user_input):
        user_input = 0 if user_input == '' else (
            int(user_input) -1 if int(user_input) > 0 else 0)
        self.directives = "%s#EndRegion\n" % self.directives
        self.window.active_view().run_command('insert_directives_au3',
            {'line': user_input, 'text': self.directives})


class InsertDirectivesAu3Command(sublime_plugin.TextCommand):
    def run(self, edit, line, text):
        point = self.view.text_point(int(line), 0)
        self.view.insert(edit, point, text)


class SyntaxCheckAu3Command(sublime_plugin.WindowCommand):
    def run(self):
        checkpath = settings.get("au3_check_path", None)
        if not checkpath:
            return

        curr_full_path = self.window.active_view().file_name()
        ext = curr_full_path[-3:]
        if not re.search("[Aa][Uu]3", ext):
            return

        checkcmd = '"%s" -q "%s"' % (checkpath, curr_full_path)
        print('\n'*10)
        timerstart = datetime.datetime.now()
        try:
            checkprocess = subprocess.Popen(
                checkcmd, shell=True, stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT, universal_newlines=True)
            timerend = datetime.datetime.now()
            checkoutput = checkprocess.communicate()[0].rstrip()
            print("------------ Beginning Au3Check ------------ %s" % time_now())
            print(checkoutput)
            sublime.active_window().run_command("show_panel", {"panel": "console"})
            if not ("error:" in checkoutput):
                print('No error')
        except Exception as e:
            sublime.active_window().run_command("show_panel", {"panel": "console"})
            print("------------ ERROR: Python exception trying to run Au3Check ------------")
            print("Au3Check Cmd was: " + " ".join(checkcmd))
            print("Error {0}".format(str(e)))

        print("-------------- Au3Check Ended -------------- %s" % time_now())


class RunWithParameterCommand(sublime_plugin.WindowCommand):
    cmd = ''
    def run(self):
        au3exe = settings.get("autoit_exe_path", None)
        if not au3exe:
            return

        curr_full_path = self.window.active_view().file_name()
        ext = curr_full_path[-3:]
        if not re.search("[Aa][Uu]3", ext):
            return

        self.cmd = '"%s" /ErrorStdOut /AutoIt3ExecuteScript "%s"' % (
            au3exe, curr_full_path)

        self.window.show_input_panel(" Insert the paramter(s) ",
            '/a /b', self.on_param, None, None)

    def on_param(self, user_input):
        if user_input == '':
            return
        else:
            self.cmd = "%s %s" % (self.cmd, user_input)
            print('\n'*10)
            print("----- Beginning Run with parameter(s) ----- %s" % time_now())
            print("cmd: %s" % self.cmd)
            self.window.run_command("exec", {"cmd": self.cmd})
            print("----------- Parameter Run Ended ----------- %s" % time_now())
            sublime.active_window().run_command("show_panel", {"panel": "console"})


class ReBuildCompletionFilesCommand(sublime_plugin.TextCommand):
    def run(self, edit):
        au3exe = settings.get("autoit_exe_path", None)
        completion_path = settings.get("completion_path", None)
        if not au3exe or not completion_path:
            return
        au3api_file = ("%sSciTE\\api\\au3.api" % au3exe[:-11])
        if not osp.exists(au3api_file):
            return

        fh = open(au3api_file)
        list_api = []
        nofunc = ''
        patt_func_params = re.compile(r'^(\w+)\s*\(\s*([^\)]+)')
        patt_nofunc = re.compile(r'^([^?]+)\?\d')
        patt_create_func_params = re.compile(
            r'^func\s+(\w+)\s*\(\s*([^\)]+)', re.I)
        while True:
            line = fh.readline()
            if not line:
                break

            detail = ''
            cursor_var = ''
            match = re.search(patt_func_params, line)
            if match:
                split_param = match.group(2).split(',')
                for i in range (0, len(split_param)):
                    split_param[i] = re.sub("\[|\]", "",
                        split_param[i]).lstrip().rstrip()
                    split_param[i] = re.sub('"', "'", split_param[i])
                    split_param[i] = re.sub(r'\$', r"\\\$", split_param[i])

                if match.group(1) in "Opt AutoItSetOption":
                    detail = " %s" % split_param[0]
                    cursor_var =  "${1:%s}" % split_param[1]
                elif match.group(1) in "ControlCommand ControlListView":
                    detail = " %s" % split_param[3]
                    n = 0
                    for i in range(0, len(split_param)):
                        if i != 3:
                            n = n + 1
                            cursor_var = "%s${%i:%s}, " % (
                                cursor_var, n, split_param[i])
                        else:
                            cursor_var = "%s%s, " % (cursor_var, split_param[i])
                else:
                    n = 0
                    for i in range(0, len(split_param)):
                        n = n + 1
                        cursor_var =  "%s${%i:%s}, " % (
                            cursor_var, n, split_param[i])

                cursor_var = re.sub(r', $', '', cursor_var)
                list_api.append(("%s%s" % (match.group(1), detail),
                    "%s(%s)" % (match.group(1), cursor_var)))
            else:
                match = re.search(patt_nofunc, line)
                if match:
                    nofunc = "%s %s" % (nofunc, match.group(1))
        fh.close()

        include_path = ("%sInclude" % au3exe[:-11])
        dict_incl = {}
        includes = ''
        if osp.exists(include_path):
            for x in os.listdir(include_path):
                if x.split(".")[1] == "au3":
                    includes = "%s <%s>" % (includes, x)

        user_include_path = settings.get("user_include_path", None)
        list_user = []
        if user_include_path:
            for x in os.listdir(user_include_path):
                if x.split(".")[1] == "au3":
                    includes = "%s <%s>" % (includes, x)
                    # read functions from this file
                    fh = open(osp.join(user_include_path, x))
                    while True:
                        line = fh.readline()
                        if not line:
                            break
                        cursor_var = ''
                        match = re.search(patt_create_func_params, line)
                        if match:
                            split_param = match.group(2).split(',')
                            n = 0
                            for i in range (0, len(split_param)):
                                n = n + 1
                                split_param[i] = re.sub("\[|\]", "",
                                    split_param[i]).lstrip().rstrip()
                                split_param[i] = re.sub('"', "'", split_param[i])
                                split_param[i] = re.sub(
                                    r'\$', r"\\\$", split_param[i])
                                cursor_var =  "%s${%i:%s}, " % (
                                    cursor_var, n, split_param[i])

                            cursor_var = re.sub(r', $', '', cursor_var)
                            list_user.append((match.group(1),
                                "%s(%s)" % (match.group(1), cursor_var)))
                    fh.close()

        if includes != '':
            dict_incl['Includes'] = includes.lstrip()

        if nofunc != '':
            dict_incl['NoFuncs'] = nofunc.lstrip()

        print('\n'*10)
        print("-- Beginning (Re)Build Completion Files --- %s" % time_now())

        sublime.active_window().run_command("show_panel", {"panel": "console"})
        for k in dict_incl.keys():
            self.write_completion_file(completion_path, k, dict_incl[k])
            print("-- ready: [ %s.sublime-completions ]" % k)

        if len(list_user) > 0:
            self.write_completion_file(completion_path, "UserUdfs", list_user, True)
            print("-- ready: [ UserUdfs.sublime-completions ]")

        if len(list_api) > 0:
            self.write_completion_file(completion_path, "Au3api", list_api, True)
            print("-- ready: [ Au3api.sublime-completions ]")

        print("------------- (Re)Build Ended ------------- %s\n" % time_now())

    def write_completion_file(self, path, filename, v, islist=False):
        file_head = '{\n"scope": "source.autoit",\n"completions": [\n'
        trigger = '    { "trigger": '
        content = ', "contents": '
        line_end = ' },\n'
        file_bottom = ']\n}\n'
        file_body = ''
        str_write = ''

        if islist:
            for i in range(0, len(v)):
                file_body = "%s%s" % (
                    file_body, '%s"%s"%s"%s"%s' % (
                        trigger,v[i][0],content,v[i][1],line_end))
        else:
            list_val = v.split(' ')
            for k in list_val:
                file_body = "%s%s" % (
                    file_body, '%s"%s"%s"%s"%s' % (trigger,k,content,k,line_end))

        str_write = "%s%s\n%s" % (file_head, file_body[:-2], file_bottom)

        new_file = osp.join(path, "%s.sublime-completions" % filename)
        if osp.exists(new_file):
            lt = localtime()
            year, month, day, hh, mm, ss = lt[0:6]
            os.rename(new_file, "%s.%04i%02i%02i%02i%02i%02i.bak" % (
                new_file, year, month, day, hh, mm, ss))
        fh = open(new_file, 'w')
        fh.write(str_write)
        fh.close()


class EventListener(sublime_plugin.EventListener):
    def on_query_completions(self, view, prefix, locations):
        ext = view.file_name()[-3:]
        if not re.search("[Aa][Uu]3", ext):
            return

        find_regions = None
        if prefix.startswith("$"):
            find_regions = view.find_all(r"\$[A-Za-z0-9_]+")
        if not find_regions:
            return

        text = set()
        for region in find_regions:
            if region == view.word(view.sel()[0]):
                continue
            text.add(view.substr(region))
        auto_complete = [(x, x.replace("$", "\\$")) for x in text]
        return (auto_complete, sublime.INHIBIT_WORD_COMPLETIONS |
            sublime.INHIBIT_EXPLICIT_COMPLETIONS)


def plugin_loaded():
    global settings
    settings = sublime.load_settings(SETTINGS_FILE)

def time_now():
    lt = localtime()
    return "[%02i:%02i:%02i]" % (lt[3], lt[4], lt[5])